import polars as pl
import logging
from ..pl_utils import to_schema, Column, DateColumn
from ..datasus import import_from_ftp
from ..utils import format_year, format_month
from ..ftp import fetch_dbc_as_df

MAIN_TABLE = "SIA_PA"


def import_sia_pa(db_file="datasus.db", years=["*"], states=["*"], months=["*"]):
    """Import PA (Produção Ambulatorial) from SIASUS (Sistema de Informações Ambulatorial do SUS).

    Args:
        db_file (str, optional): path to the duckdb file in which the data will be imported to. Defaults to "datasus.db".
        years (list, optional): list of years for which data will be imported (if available). Eg: `[2012, 2000, 2010]`. Defaults to ["*"].
        states (list, optional): list of brazilian 2 letters state for which data will be imported (if available). Eg: `["SP", "RJ"]`. Defaults to ["*"].
        months (list, optional): list of months numbers (1-12) for which data will be imported (if available). Eg: `[1, 12, 6]`. Defaults to ["*"].

    ---

    Extra:
    - **Data description**: https://github.com/mymatsubara/datasus-db/blob/main/docs/sia_pa.pdf
    - **ftp path**: ftp.datasus.gov.br/dissemin/publicos/SIASUS/200801_/Dados/PA*.dbc
    """
    logging.info(f"⏳ [{MAIN_TABLE}] Starting import...")
    import_from_ftp(
        [MAIN_TABLE],
        [
            f"/dissemin/publicos/SIASUS/200801_/Dados/PA{state.upper()}{format_year(year)}{format_month(month)}*.dbc"
            for year in years
            for state in states
            for month in months
        ],
        fetch_sia_rh,
        db_file=db_file,
    )


def fetch_sia_rh(ftp_path: str):
    df = fetch_dbc_as_df(ftp_path)
    return {MAIN_TABLE: map_sia_pa(df)}


def map_sia_pa(df: pl.DataFrame):
    df = df.with_columns(
        pl.when(pl.col(pl.Utf8).str.len_chars() == 0)
        .then(None)
        .otherwise(pl.col(pl.Utf8))
        .name.keep(),
    )

    return to_schema(
        df,
        [
            Column("PA_CODUNI", pl.Utf8),
            Column("PA_GESTAO", pl.Utf8),
            Column("PA_CONDIC", pl.Utf8),
            Column("PA_UFMUN", pl.Utf8),
            Column("PA_REGCT", pl.Utf8),
            Column("PA_INCOUT", pl.Utf8),
            Column("PA_INCURG", pl.Utf8),
            Column("PA_TPUPS", pl.Utf8),
            Column("PA_TIPPRE", pl.Utf8),
            Column("PA_MN_IND", pl.Utf8),
            Column("PA_CNPJCPF", pl.Utf8),
            Column("PA_CNPJMNT", pl.Utf8),
            Column("PA_CNPJ_CC", pl.Utf8),
            Column("PA_MVM", pl.Utf8),
            Column("PA_CMP", pl.Utf8),
            Column("PA_PROC_ID", pl.Utf8),
            Column("PA_TPFIN", pl.Utf8),
            Column("PA_SUBFIN", pl.Utf8),
            Column("PA_NIVCPL", pl.Utf8),
            Column("PA_DOCORIG", pl.Utf8),
            Column("PA_AUTORIZ", pl.Utf8),
            Column("PA_CNSMED", pl.Utf8),
            Column("PA_CBOCOD", pl.Utf8),
            Column("PA_MOTSAI", pl.Utf8),
            Column("PA_OBITO", pl.Utf8),
            Column("PA_ENCERR", pl.Utf8),
            Column("PA_PERMAN", pl.Utf8),
            Column("PA_ALTA", pl.Utf8),
            Column("PA_TRANSF", pl.Utf8),
            Column("PA_CIDPRI", pl.Utf8),
            Column("PA_CIDSEC", pl.Utf8),
            Column("PA_CIDCAS", pl.Utf8),
            Column("PA_CATEND", pl.Utf8),
            Column("PA_IDADE", pl.Utf8),
            Column("IDADEMIN", pl.Utf8),
            Column("IDADEMAX", pl.Utf8),
            Column("PA_FLIDADE", pl.Utf8),
            Column("PA_SEXO", pl.Utf8),
            Column("PA_RACACOR", pl.Utf8),
            Column("PA_MUNPCN", pl.Utf8),
            Column("PA_QTDPRO", pl.Utf8),
            Column("PA_QTDAPR", pl.Utf8),
            Column("PA_VALPRO", pl.Utf8),
            Column("PA_VALAPR", pl.Utf8),
            Column("PA_UFDIF", pl.Utf8),
            Column("PA_MNDIF", pl.Utf8),
            Column("PA_DIF_VAL", pl.Utf8),
            Column("NU_VPA_TOT", pl.Utf8),
            Column("NU_PA_TOT", pl.Utf8),
            Column("PA_INDICA", pl.Utf8),
            Column("PA_CODOCO", pl.Utf8),
            Column("PA_FLQT", pl.Utf8),
            Column("PA_FLER", pl.Utf8),
            Column("PA_ETNIA", pl.Utf8),
            Column("PA_VL_CF", pl.Float64),
            Column("PA_VL_CL", pl.Float64),
            Column("PA_VL_INC", pl.Float64),
            Column("PA_SRV_C", pl.Utf8),
            Column("PA_INE", pl.Utf8),
            Column("PA_NAT_JUR", pl.Utf8),
        ],
    )



"""
Module with helper functions to interact with DATASUS ftp server
"""

import urllib.request as request
import ftplib
import logging
import io
from typing import Iterable
import os.path as path
import os
from zipfile import ZipFile
from dbfread import DBF
import polars as pl
import datasus_dbc
from .utils import rm, flatten


def fetch_dbc_as_df(ftp_path: str) -> pl.DataFrame:
    response = request.urlopen(ftp_path)
    dbc_raw = response.read()

    filename = path.basename(ftp_path).split(".")[0]
    dbc_file = f".tmp/{filename}.dbc"
    dbf_file = f".tmp/{filename}.dbf"

    os.makedirs(path.dirname(dbc_file), exist_ok=True)
    with open(
        dbc_file,
        "wb",
    ) as f:
        f.write(dbc_raw)

    datasus_dbc.decompress(dbc_file, dbf_file)

    df = pl.DataFrame(iter(DBF(dbf_file, encoding="iso-8859-1")))

    rm(dbc_file)
    rm(dbf_file)

    return df


def get_matching_files(host: str, patterns: Iterable[str]):
    ftp = ftplib.FTP(host)
    ftp.login()

    return set(flatten((try_nlst(pattern, ftp) for pattern in patterns)))


def try_nlst(pattern: str, ftp: ftplib.FTP):
    files = ftp.nlst(pattern)
    if len(files) == 0:
        logging.warn(f"⚠️  Could not found file matching: {pattern}")

    return files


def fetch_from_zip(ftp_path: str, files: list[str]):
    response = request.urlopen(ftp_path)
    zip_file = ZipFile(io.BytesIO(response.read()))

    lowercase_filenames = {
        file.filename.lower(): file.filename for file in zip_file.filelist
    }

    return {file: zip_file.read(lowercase_filenames[file.lower()]) for file in files}



import polars as pl
import os.path as path
import logging
from ..pl_utils import to_schema, Column
from ..dbf import read_as_df
from ..datasus import import_from_ftp
from ..utils import format_year
from ..ftp import fetch_from_zip

MAIN_TABLE = "IBGE_POP_TCU"


def import_ibge_pop_tcu(db_file="datasus.db", years=["*"]):
    """Import population estimated per city by TCU (Tribunal de Contas da União).

    Args:
        db_file (str, optional): path to the duckdb file in which the data will be imported to. Defaults to "datasus.db".
        years (list, optional): list of years for which data will be imported (if available). Eg: `[2012, 2000, 2010]`. Defaults to ["*"].

    ---

    Extra:
    - **Data description**: https://github.com/mymatsubara/datasus-db/blob/main/docs/ibge_pop_tcu.pdf
    - **ftp path**: ftp.datasus.gov.br/dissemin/publicos/IBGE/POPTCU/POPTBR*.zip
    """
    logging.info(f"⏳ [{MAIN_TABLE}] Starting import...")

    import_from_ftp(
        [MAIN_TABLE],
        [
            f"/dissemin/publicos/IBGE/POPTCU/POPTBR{format_year(year)}.zip*"
            for year in years
        ],
        fetch_ibge_pop_tcu,
        db_file,
    )


def fetch_ibge_pop_tcu(ftp_path: str):
    dbf_file = path.basename(ftp_path).split(".")[0] + ".dbf"
    files = fetch_from_zip(ftp_path, [dbf_file])

    df = read_as_df(dbf_file, files[dbf_file])

    return {MAIN_TABLE: map_ibge_pop_tcu(df)}


def map_ibge_pop_tcu(df: pl.DataFrame):
    return to_schema(
        df,
        [
            Column("MUNIC_RES", pl.UInt32),
            Column("ANO", pl.UInt16),
            Column("POPULACAO", pl.UInt32),
        ],
    ).with_columns(
        pl.when(pl.col("MUNIC_RES") >= 1_000_000)
        .then(pl.col("MUNIC_RES") // 10)
        .otherwise(pl.col("MUNIC_RES"))
        .name.keep()
    )



import polars as pl
import logging
from ..pl_utils import to_schema, Column, DateColumn
from ..datasus import import_from_ftp
from ..utils import format_year, format_month
from ..ftp import fetch_dbc_as_df

MAIN_TABLE = "SIH_SP"


def import_sih_sp(db_file="datasus.db", years=["*"], states=["*"], months=["*"]):
    """Import SP (Autorização de Internação Hospitalar Saúde do Profissional) from SIHSUS (Sistema de Informações Hospitalares do SUS).

    Args:
        db_file (str, optional): path to the duckdb file in which the data will be imported to. Defaults to "datasus.db".
        years (list, optional): list of years for which data will be imported (if available). Eg: `[2012, 2000, 2010]`. Defaults to ["*"].
        states (list, optional): list of brazilian 2 letters state for which data will be imported (if available). Eg: `["SP", "RJ"]`. Defaults to ["*"].
        months (list, optional): list of months numbers (1-12) for which data will be imported (if available). Eg: `[1, 12, 6]`. Defaults to ["*"].

    ---

    Extra:
    - **Data description**: https://github.com/mymatsubara/datasus-db/blob/main/docs/sih_sp.pdf
    - **ftp path**: ftp.datasus.gov.br/dissemin/publicos/SIHSUS/200801_/Dados/SP*.dbc
    """
    logging.info(f"⏳ [{MAIN_TABLE}] Starting import...")

    import_from_ftp(
        [MAIN_TABLE],
        [
            f"/dissemin/publicos/SIHSUS/200801_/Dados/SP{state.upper()}{format_year(year)}{format_month(month)}.dbc*"
            for year in years
            for state in states
            for month in months
        ],
        fetch_sih_rh,
        db_file=db_file,
    )


def fetch_sih_rh(ftp_path: str):
    df = fetch_dbc_as_df(ftp_path)
    return {MAIN_TABLE: map_sih_sp(df)}


def map_sih_sp(df: pl.DataFrame):
    df = df.with_columns(
        pl.when(pl.col(pl.Utf8).str.len_chars() == 0)
        .then(None)
        .otherwise(pl.col(pl.Utf8))
        .name.keep(),
    )

    return to_schema(
        df,
        [
            Column("SP_GESTOR", pl.Utf8),
            Column("SP_UF", pl.Utf8),
            Column("SP_AA", pl.Utf8),
            Column("SP_MM", pl.Utf8),
            Column("SP_CNES", pl.Utf8),
            Column("SP_NAIH", pl.Utf8),
            Column("SP_PROCREA", pl.Utf8),
            Column("SP_DTINTER", pl.Utf8),
            Column("SP_DTSAIDA", pl.Utf8),
            Column("SP_NUM_PR", pl.Utf8),
            Column("SP_TIPO", pl.Utf8),
            Column("SP_CPFCGC", pl.Utf8),
            Column("SP_ATOPROF", pl.Utf8),
            Column("SP_TP_ATO", pl.Utf8),
            Column("SP_QTD_ATO", pl.Utf8),
            Column("SP_PTSP", pl.Utf8),
            Column("SP_NF", pl.Utf8),
            Column("SP_VALATO", pl.Utf8),
            Column("SP_M_HOSP", pl.Utf8),
            Column("SP_M_PAC", pl.Utf8),
            Column("SP_DES_HOS", pl.Utf8),
            Column("SP_DES_PAC", pl.Utf8),
            Column("SP_COMPLEX", pl.Utf8),
            Column("SP_FINANC", pl.Utf8),
            Column("SP_CO_FAEC", pl.Utf8),
            Column("SP_PF_CBO", pl.Utf8),
            Column("SP_PF_DOC", pl.Utf8),
            Column("SP_PJ_DOC", pl.Utf8),
            Column("IN_TP_VAL", pl.Utf8),
            Column("SEQUENCIA", pl.Utf8),
            Column("REMESSA", pl.Utf8),
            Column("SERV_CLA", pl.Utf8),
            Column("SP_CIDPRI", pl.Utf8),
            Column("SP_CIDSEC", pl.Utf8),
            Column("SP_QT_PROC", pl.Utf8),
            Column("SP_U_AIH", pl.Utf8),
        ],
    )



"""
Module with functions used to batch multiple imports from DATASUS's ftp server in parallel 
"""

from typing import Callable
import os.path as path
import duckdb
import multiprocessing
import polars as pl
import time
import random
import re
import logging
from typing import Iterable
from .ftp import get_matching_files
from .db import (
    create_import_table,
    check_new_files,
    import_dataframe,
    is_file_imported,
    mark_file_as_imported,
)


MapFn = Callable[[pl.DataFrame], pl.DataFrame]
FetchFn = Callable[[str], dict[str, pl.DataFrame]]


def import_from_ftp(
    target_tables: list[str],
    ftp_globs: Iterable[str],
    fetch_fn: FetchFn,
    db_file="datasus.db",
    ftp_host="ftp.datasus.gov.br",
    ftp_exclude_regex: str = None,
):
    with duckdb.connect(db_file) as db_con:
        target_tables_set = set(target_tables)
        files = get_matching_files(ftp_host, ftp_globs)
        if ftp_exclude_regex:
            files = remove_matching(files, ftp_exclude_regex)

        create_import_table(db_con)
        new_files = check_new_files(files, target_tables, db_con)
        new_filepaths = [f"ftp://{ftp_host}{file}" for file in new_files]

        # Shuffle files to import in random order to reduce the chance of importing multiple large files at the same time
        random.shuffle(new_filepaths)

        # Fetch dataframes in parallel
        processes_count = max(min(multiprocessing.cpu_count(), len(new_filepaths)), 1)
        total_files = len(new_filepaths)
        files_imported = 0
        errors: list[tuple[str, Exception]] = []

        # Batching is done to make sure the garbage collector kicks in
        for new_filepaths in batch(new_filepaths, 64):
            with multiprocessing.Pool(processes=processes_count) as pool:
                waiting = [
                    (
                        filepath,
                        pool.apply_async(
                            log_fetch,
                            args=(filepath, fetch_fn, logging.getLogger().level),
                        ),
                    )
                    for filepath in new_filepaths
                ]

                while len(waiting) != 0:
                    still_wating = []

                    for filepath, process in waiting:
                        if process.ready():
                            try:
                                # Import fetched data
                                filename = path.basename(filepath)
                                tables_data = process.get()

                                msg = f"📂 [{files_imported + 1}/{total_files}] Importing data from file {filename}"
                                logging.info(msg)

                                for table in tables_data.keys():
                                    if not table in target_tables_set:
                                        logging.error(
                                            f"❌ Table name '{table}' not declared in 'target_tables': {target_tables}"
                                        )
                                        continue

                                    if is_file_imported(filename, table, db_con):
                                        msg = f"🗃️ [{table}] File '{filename}' already imported"
                                        logging.info(msg)
                                        continue

                                    df = tables_data[table]
                                    import_table_data(df, table, filepath, db_con)

                            except Exception as e:
                                logging.error(f"❌ Error while importing '{filepath}'")
                                logging.error("Message: ", e)
                                errors.append((filepath, e))

                            files_imported += 1

                        else:
                            still_wating.append((filepath, process))

                    waiting = still_wating
                    time.sleep(0.5)

    if len(errors) == 0:
        logging.info(f"✅ Data successfully imported to tables: {target_tables}")
    else:
        logging.error(
            f"⚠️  {len(errors)} out of {total_files} imports failed:",
        )
        for filepath, e in errors:
            logging.error(f"    ❌ {path.basename(filepath)}: {e}")


def log_fetch(ftp_path: str, fetch_fn: FetchFn, log_level: int):
    logging.getLogger().setLevel(log_level)
    logging.info(f"⬇️  Downloading file from ftp: '{ftp_path}'")
    return fetch_fn(ftp_path)


def batch(iterable, n=1):
    l = len(iterable)
    for ndx in range(0, l, n):
        yield iterable[ndx : min(ndx + n, l)]


def import_table_data(
    df: pl.DataFrame,
    target_table: str,
    filepath: str,
    db_con: duckdb.DuckDBPyConnection,
):
    filename = path.basename(filepath)
    logging.info(f"💾 [{target_table}] Saving data to database from: {filename}")
    row_count = df.select(pl.count())[0, 0]

    if row_count != 0:
        import_dataframe(target_table, df, db_con)
    else:
        logging.warning(f"⚠️ [{target_table}] '{filename}' has no data")

    mark_file_as_imported(filepath, target_table, db_con)


def remove_matching(list: list[str], regex: str):
    compiled = re.compile(regex)
    return [e for e in list if not compiled.match(e)]



"""
[![PyPI version](https://badge.fury.io/py/datasus-db.svg)](https://pypi.org/project/datasus-db/)

A python package to **download and import** public available data from **DATASUS's** ftp servers into a [DuckDB](https://duckdb.org/) database.


# Import functions
Bellow is the list of all **import functions**:
- `datasus_db.datasources.sih_rd.import_sih_rd`
- `datasus_db.datasources.sih_sp.import_sih_sp`
- `datasus_db.datasources.sim_do.import_sim_do`
- `datasus_db.datasources.sia_pa.import_sia_pa`
- `datasus_db.datasources.po.import_po`
- `datasus_db.datasources.ibge_pop.import_ibge_pop`
- `datasus_db.datasources.ibge_pop_tcu.import_ibge_pop_tcu`
- `datasus_db.datasources.auxiliar.import_auxiliar_tables`


## Datasources
The list of all available DATASUS's datasources can be seen here: https://datasus.saude.gov.br/transferencia-de-arquivos/

If `datasus_db` is missing a datasource that you need, feel free to create an issue here: https://github.com/mymatsubara/datasus-db/issues/new
"""

from .datasources.sih_rd import import_sih_rd
from .datasources.sih_sp import import_sih_sp
from .datasources.sim_do import import_sim_do
from .datasources.sia_pa import import_sia_pa
from .datasources.po import import_po
from .datasources.ibge_pop import import_ibge_pop
from .datasources.ibge_pop_tcu import import_ibge_pop_tcu
from .datasources.auxiliar import import_auxiliar_tables



import duckdb

VIEW_DDL = """
CREATE OR REPLACE VIEW IBGE_PIRAMIDE_ETARIA AS with POP AS (
	select 
		MUNIC_RES, 
		ANO, 
		SEXO, 
		INICIO_FXETARIA, 
		FIM_FXETARIA, 
		SUM(POPULACAO) AS POPULACAO 
	from main.IBGE_POP
	group by MUNIC_RES, ANO, SEXO, INICIO_FXETARIA, FIM_FXETARIA
),
POP_PERCENTUAL AS (
	select 
		MUNIC_RES,
		SEXO,
		ANO,
		INICIO_FXETARIA,
		FIM_FXETARIA,
		POPULACAO,
		POPULACAO / SUM(POPULACAO) OVER (PARTITION BY MUNIC_RES, ANO) AS PERCETUAL,
	from POP
)
select MUNIC_RES, SEXO, ANO, INICIO_FXETARIA, FIM_FXETARIA, POPULACAO, PERCETUAL 
from POP_PERCENTUAL;
"""


def create_piramide_etaria_view(db_file="datasus.db"):
    with duckdb.connect(db_file) as con:
        con.execute(VIEW_DDL)



import logging
import polars as pl
from ..pl_utils import fill_empty, Column, to_schema
from ..ftp import fetch_from_zip
from ..cnv import to_dataframe
from ..datasus import import_from_ftp
from ..dbf import read_as_df


MUNICIPIO_TABLE = "AUX_MUNICIPIO"
UF_TABLE = "AUX_UF"
CID10_DOENCA_TABLE = "AUX_CID10_DOENCA"


def import_auxiliar_tables(db_file="datasus.db"):
    """Import auxiliar tables with some datasus codes definitions (eg: municipios, doenças, ...)

    Args:
        db_file (str, optional): path to the duckdb file in which the data will be imported to.

    ---

    Extra:
    - **Municipio data description**: https://github.com/mymatsubara/datasus-db/blob/main/docs/auxiliar/municipio.pdf
    - **ftp path**: ftp.datasus.gov.br/dissemin/publicos/SIM/CID10/DOCS/Docs_Tabs_CID10.zip
    """
    logging.info(f"⏳ [AUX_TABLES] Starting import...")

    import_from_ftp(
        [CID10_DOENCA_TABLE, MUNICIPIO_TABLE, UF_TABLE],
        ["/dissemin/publicos/SIM/CID10/DOCS/Docs_Tabs_CID10.zip*"],
        fetch_sim_auxiliar,
        db_file=db_file,
    )


def fetch_sim_auxiliar(ftp_path: str):
    cid10_file = "TABELAS/CID10.DBF"
    municipio_file = "TABELAS/CADMUN.DBF"
    uf_file = "TABELAS/TABUF.DBF"
    files = fetch_from_zip(ftp_path, [cid10_file, municipio_file, uf_file])

    cid10_df = read_as_df(cid10_file, files[cid10_file], encoding="cp850")
    municipio_df = read_as_df(municipio_file, files[municipio_file], encoding="cp850")
    uf_df = read_as_df(uf_file, files[uf_file], encoding="cp850")

    return {
        CID10_DOENCA_TABLE: map_cid10(cid10_df),
        MUNICIPIO_TABLE: map_municipio(municipio_df),
        UF_TABLE: map_uf(uf_df),
    }


def map_cid10(df: pl.DataFrame):
    df = df.with_columns(fill_empty(None))

    return to_schema(
        df,
        [
            Column("CID10", pl.Utf8),
            Column("OPC", pl.Utf8),
            Column("CAT", pl.Utf8),
            Column("SUBCAT", pl.Utf8),
            Column("DESCR", pl.Utf8),
            Column("RESTRSEXO", pl.UInt8),
        ],
    )


def map_municipio(df: pl.DataFrame):
    df = df.with_columns(fill_empty(None))

    return to_schema(
        df,
        [
            Column("MUNCOD", pl.UInt32),
            Column("MUNCODDV", pl.UInt32),
            Column("SITUACAO", pl.Utf8),
            Column("MUNSINP", pl.UInt32),
            Column("MUNSIAFI", pl.UInt32),
            Column("MUNNOME", pl.Utf8),
            Column("MUNNOMEX", pl.Utf8),
            Column("OBSERV", pl.Utf8),
            Column("MUNSINON", pl.Utf8),
            Column("MUNSINONDV", pl.Utf8),
            Column("AMAZONIA", pl.Utf8),
            Column("FRONTEIRA", pl.Utf8),
            Column("CAPITAL", pl.Utf8),
            Column("UFCOD", pl.UInt8),
            Column("MESOCOD", pl.UInt16),
            Column("MICROCOD", pl.UInt16),
            Column("MSAUDCOD", pl.UInt16),
            Column("RSAUDCOD", pl.UInt16),
            Column("CSAUDCOD", pl.UInt16),
            Column("RMETRCOD", pl.UInt16),
            Column("AGLCOD", pl.UInt16),
            Column("ANOINST", pl.UInt16),
            Column("ANOEXT", pl.UInt16),
            Column("SUCESSOR", pl.UInt32),
            Column("LATITUDE", pl.Float64),
            Column("LONGITUDE", pl.Float64),
            Column("ALTITUDE", pl.Float64),
            Column("AREA", pl.Float64),
        ],
    )


def map_uf(df: pl.DataFrame):
    return to_schema(
        df,
        [
            Column("SIGLA_UF", pl.Utf8),
            Column("CODIGO", pl.UInt8),
            Column("DESCRICAO", pl.Utf8),
        ],
    )


def fetch_painel_oncologia_auxiliar(ftp_path: str):
    municipio_file = "CNV/br_municip.cnv"
    uf_file = "CNV/br_uf.cnv"

    files = fetch_from_zip(ftp_path, [municipio_file, uf_file])

    def read_as_df(file_name: str, id_dtype: pl.UInt32):
        cnv_bytes = files[file_name]
        df = to_dataframe(cnv_bytes, id_dtype=id_dtype)
        return df.with_columns(
            pl.col("NOME").str.split(" ").list.slice(1).list.join(" ")
        )

    return {
        MUNICIPIO_TABLE: read_as_df(municipio_file, id_dtype=pl.UInt32),
        UF_TABLE: read_as_df(uf_file, id_dtype=pl.UInt8),
    }



"""
Module with helper functions to handler with *.dbf files
"""
import os.path as path
import polars as pl
from dbfread import DBF
from .utils import rm


def read_as_df(filename: str, bytes: bytes, encoding: str = None):
    tmp_file = path.join(".tmp", path.basename(filename).split(".")[0] + ".dbf")

    with open(tmp_file, "wb") as f:
        f.write(bytes)

    try:
        dbf = DBF(tmp_file, encoding=encoding)
        df = pl.DataFrame(iter(dbf))
        return df
    except Exception as e:
        raise e
    finally:
        rm(tmp_file)



"""Module with generic helper functions"""

import itertools
import os


def format_year(year, digits=2):
    if isinstance(year, str) and not year.isnumeric():
        return "*"

    year = str(year).zfill(digits)

    return year[-digits:]


def format_month(month):
    if isinstance(month, str) and not month.isnumeric():
        return "*"

    return str(month).zfill(2)


def flatten(list_of_lists):
    return itertools.chain.from_iterable(list_of_lists)


def rm(file: str):
    if os.path.exists(file):
        os.remove(file)



import polars as pl
import logging
from ..pl_utils import to_schema, Column, DateColumn
from ..datasus import import_from_ftp
from ..utils import format_year, format_month
from ..ftp import fetch_dbc_as_df

MAIN_TABLE = "SIH_RD"


def import_sih_rd(db_file="datasus.db", years=["*"], states=["*"], months=["*"]):
    """Import RD (Autorização de Internação Hospitalar Reduzida) from SIMSUS (Sistema de Informações Hospitalares do SUS).

    Args:
        db_file (str, optional): path to the duckdb file in which the data will be imported to. Defaults to "datasus.db".
        years (list, optional): list of years for which data will be imported (if available). Eg: `[2012, 2000, 2010]`. Defaults to ["*"].
        states (list, optional): list of brazilian 2 letters state for which data will be imported (if available). Eg: `["SP", "RJ"]`. Defaults to ["*"].
        months (list, optional): list of months numbers (1-12) for which data will be imported (if available). Eg: `[1, 12, 6]`. Defaults to ["*"].

    ---

    Extra:
    - **Data description**: https://github.com/mymatsubara/datasus-db/blob/main/docs/sih_rd.pdf
    - **ftp path**: ftp.datasus.gov.br/dissemin/publicos/SIHSUS/200801_/Dados/RD*.dbc
    """
    logging.info(f"⏳ [{MAIN_TABLE}] Starting import...")

    import_from_ftp(
        [MAIN_TABLE],
        [
            f"/dissemin/publicos/SIHSUS/200801_/Dados/RD{state.upper()}{format_year(year)}{format_month(month)}.dbc*"
            for year in years
            for state in states
            for month in months
        ],
        fetch_sih_rh,
        db_file=db_file,
    )


def fetch_sih_rh(ftp_path: str):
    df = fetch_dbc_as_df(ftp_path)
    return {MAIN_TABLE: map_sih_rd(df)}


def map_sih_rd(df: pl.DataFrame):
    df = df.with_columns(
        pl.when(pl.col(pl.Utf8).str.len_chars() == 0)
        .then(None)
        .otherwise(pl.col(pl.Utf8))
        .name.keep(),
    ).with_columns(
        pl.when(pl.col("GESTOR_CPF").str.contains("[1-9]"))
        .then(pl.col("GESTOR_CPF"))
        .otherwise(None)
        .name.keep(),
        pl.when(pl.col("INSC_PN").str.contains("[1-9]"))
        .then(pl.col("INSC_PN"))
        .otherwise(None)
        .name.keep(),
    )

    return to_schema(
        df,
        [
            Column("UF_ZI", pl.Utf8),
            Column("ANO_CMPT", pl.Utf8),
            Column("MES_CMPT", pl.Utf8),
            Column("ESPEC", pl.Utf8),
            Column("CGC_HOSP", pl.Utf8),
            Column("N_AIH", pl.Utf8),
            Column("IDENT", pl.Utf8),
            Column("CEP", pl.Utf8),
            Column("MUNIC_RES", pl.Utf8),
            Column("NASC", pl.Utf8),
            Column("SEXO", pl.Utf8),
            Column("UTI_MES_IN", pl.Utf8),
            Column("UTI_MES_AN", pl.Utf8),
            Column("UTI_MES_AL", pl.Utf8),
            Column("UTI_MES_TO", pl.Utf8),
            Column("UTI_INT_IN", pl.Utf8),
            Column("UTI_INT_AN", pl.Utf8),
            Column("UTI_INT_AL", pl.Utf8),
            Column("UTI_INT_TO", pl.Utf8),
            Column("DIAR_ACOM", pl.Utf8),
            Column("QT_DIARIAS", pl.Utf8),
            Column("PROC_SOLIC", pl.Utf8),
            Column("PROC_REA", pl.Utf8),
            Column("VAL_SH", pl.Float64),
            Column("VAL_SP", pl.Float64),
            Column("VAL_SADT", pl.Float64),
            Column("VAL_RN", pl.Float64),
            Column("VAL_ACOMP", pl.Float64),
            Column("VAL_ORTP", pl.Float64),
            Column("VAL_SANGUE", pl.Float64),
            Column("VAL_SADTSR", pl.Float64),
            Column("VAL_TRANSP", pl.Float64),
            Column("VAL_OBSANG", pl.Float64),
            Column("VAL_PED1AC", pl.Float64),
            Column("VAL_TOT", pl.Float64),
            Column("VAL_UTI", pl.Float64),
            Column("US_TOT", pl.Float64),
            Column("DT_INTER", pl.Utf8),
            Column("DT_SAIDA", pl.Utf8),
            Column("DIAG_PRINC", pl.Utf8),
            Column("DIAG_SECUN", pl.Utf8),
            Column("COBRANCA", pl.Utf8),
            Column("NATUREZA", pl.Utf8),
            Column("NAT_JUR", pl.Utf8),
            Column("GESTAO", pl.Utf8),
            Column("RUBRICA", pl.Utf8),
            Column("IND_VDRL", pl.Utf8),
            Column("MUNIC_MOV", pl.Utf8),
            Column("COD_IDADE", pl.Utf8),
            Column("IDADE", pl.Utf8, strict=False),
            Column("DIAS_PERM", pl.Utf8),
            Column("MORTE", pl.Utf8),
            Column("NACIONAL", pl.Utf8),
            Column("NUM_PROC", pl.Utf8),
            Column("CAR_INT", pl.Utf8),
            Column("TOT_PT_SP", pl.Utf8),
            Column("CPF_AUT", pl.Utf8),
            Column("HOMONIMO", pl.Utf8),
            Column("NUM_FILHOS", pl.Utf8),
            Column("INSTRU", pl.Utf8),
            Column("CID_NOTIF", pl.Utf8),
            Column("CONTRACEP1", pl.Utf8),
            Column("CONTRACEP2", pl.Utf8),
            Column("GESTRISCO", pl.Utf8),
            Column("INSC_PN", pl.Utf8),
            Column("SEQ_AIH5", pl.Utf8),
            Column("CBOR", pl.Utf8),
            Column("CNAER", pl.Utf8),
            Column("VINCPREV", pl.Utf8),
            Column("GESTOR_COD", pl.Utf8),
            Column("GESTOR_TP", pl.Utf8),
            Column("GESTOR_CPF", pl.Utf8),
            Column("GESTOR_DT", pl.Utf8),
            Column("CNES", pl.Utf8),
            Column("CNPJ_MANT", pl.Utf8),
            Column("INFEHOSP", pl.Utf8),
            Column("CID_ASSO", pl.Utf8),
            Column("CID_MORTE", pl.Utf8),
            Column("COMPLEX", pl.UInt8),
            Column("FINANC", pl.UInt8),
            Column("FAEC_TP", pl.Utf8),
            Column("REGCT", pl.Utf8),
            Column("RACA_COR", pl.UInt8),
            Column("ETNIA", pl.Utf8),
            Column("SEQUENCIA", pl.Utf8),
            Column("REMESSA", pl.Utf8),
            Column("AUD_JUST", pl.Utf8),
            Column("SIS_JUST", pl.Utf8),
            Column("VAL_SH_FED", pl.Float64),
            Column("VAL_SP_FED", pl.Float64),
            Column("VAL_SH_GES", pl.Float64),
            Column("VAL_SP_GES", pl.Float64),
            Column("VAL_UCI", pl.Float64),
            Column("MARCA_UCI", pl.UInt8),
            Column("DIAGSEC1", pl.Utf8),
            Column("DIAGSEC2", pl.Utf8),
            Column("DIAGSEC3", pl.Utf8),
            Column("DIAGSEC4", pl.Utf8),
            Column("DIAGSEC5", pl.Utf8),
            Column("DIAGSEC6", pl.Utf8),
            Column("DIAGSEC7", pl.Utf8),
            Column("DIAGSEC8", pl.Utf8),
            Column("DIAGSEC9", pl.Utf8),
            Column("TPDISEC1", pl.UInt8),
            Column("TPDISEC2", pl.UInt8),
            Column("TPDISEC3", pl.UInt8),
            Column("TPDISEC4", pl.UInt8),
            Column("TPDISEC5", pl.UInt8),
            Column("TPDISEC6", pl.UInt8),
            Column("TPDISEC7", pl.UInt8),
            Column("TPDISEC8", pl.UInt8),
            Column("TPDISEC9", pl.UInt8),
        ],
    )



import polars as pl
import logging
from ..pl_utils import (
    upsert_column,
    to_schema,
    Column,
    DateColumn,
    rename_columns,
    fill_empty,
    fill_text,
    fill_non_numeric,
)
from ..datasus import import_from_ftp
from ..utils import format_year
from ..ftp import fetch_dbc_as_df

MAIN_TABLE = "SIM_DO"


def import_sim_do(db_file="datasus.db", years=["*"], states=["*"]):
    """Import DO (Declaração de Óbito) from SIM (Sistema de informações de Mortalidade).

    Args:
        db_file (str, optional): path to the duckdb file in which the data will be imported to. Defaults to "datasus.db".
        years (list, optional): list of years for which data will be imported (if available). Eg: `[2012, 2000, 2010]`. Defaults to ["*"].
        states (list, optional): list of brazilian 2 letters state for which data will be imported (if available). Eg: `["SP", "RJ"]`. Defaults to ["*"].

    ---

    Extra:
    - **Data description**: https://github.com/mymatsubara/datasus-db/blob/main/docs/sim_do.pdf
    - **ftp path non preliminary data**: ftp.datasus.gov.br/dissemin/publicos/SIM/CID10/DORES/DO*.dbc
    - **ftp path preliminary data**: ftp.datasus.gov.br/dissemin/publicos/SIM/PRELIM/DORES/DO*.dbc
    """
    logging.info(f"⏳ [{MAIN_TABLE}] Starting import for non preliminary data...")
    import_from_ftp(
        [MAIN_TABLE],
        [
            f"/dissemin/publicos/SIM/CID10/DORES/DO{state.upper()}{format_year(year, digits=4)}.dbc*"
            for year in years
            for state in states
        ],
        fetch_sim_do,
        ftp_exclude_regex=r".*/DOBR.*\.dbc",
        db_file=db_file,
    )

    logging.info(f"⏳ [{MAIN_TABLE}] Starting import for preliminary data...")
    import_from_ftp(
        [MAIN_TABLE],
        [
            f"/dissemin/publicos/SIM/PRELIM/DORES/DO{state.upper()}{format_year(year, digits=4)}.dbc*"
            for year in years
            for state in states
        ],
        fetch_sim_do,
        ftp_exclude_regex=r".*/DOBR.*\.dbc",
        db_file=db_file,
    )


def fetch_sim_do(ftp_path: str):
    df = fetch_dbc_as_df(ftp_path)
    return {MAIN_TABLE: map_sim_do(df)}


def map_sim_do(df: pl.DataFrame):
    df = (
        df.with_columns(fill_empty(None))
        .with_columns(
            fill_text("NULL", None),
        )
        .with_columns(
            fill_text("00000000", None),
        )
        .with_columns(
            fill_non_numeric(None, pl.col(["PESO", "NATURAL"])),
        )
    )
    df = rename_columns(df, {"contador": "CONTADOR"})

    df = (
        df.with_columns(upsert_column(df, "DTCADASTRO", pl.Utf8))
        .with_columns(
            pl.when(pl.col("DTOBITO").str.len_chars() == 4)
            .then("0101" + pl.col("DTOBITO"))
            .otherwise(pl.col("DTOBITO"))
            .name.keep(),
            pl.when(pl.col("DTNASC").str.len_chars() == 4)
            .then("0101" + pl.col("DTNASC"))
            .otherwise(pl.col("DTNASC"))
            .name.keep(),
            pl.when(pl.col("DTCADASTRO").str.len_chars() == 7)
            .then("0" + pl.col("DTCADASTRO"))
            .otherwise(pl.col("DTCADASTRO"))
            .name.keep(),
        )
        .with_columns(
            pl.when(pl.col("DTOBITO").str.len_chars() == 6)
            .then("01" + pl.col("DTOBITO"))
            .otherwise(pl.col("DTOBITO"))
            .name.keep(),
            pl.when(pl.col("DTNASC").str.len_chars() == 6)
            .then("01" + pl.col("DTNASC"))
            .otherwise(pl.col("DTNASC"))
            .name.keep(),
        )
    )

    return to_schema(
        df,
        [
            Column("ORIGEM", pl.UInt8),
            Column("TIPOBITO", pl.UInt8),
            DateColumn("DTOBITO", "%d%m%Y", strict=False),
            Column("HORAOBITO", pl.Utf8),
            Column("NATURAL", pl.UInt32),
            Column("CODMUNNATU", pl.UInt32),
            DateColumn("DTNASC", "%d%m%Y", strict=False),
            Column("IDADE", pl.UInt16),
            Column("SEXO", pl.UInt8),
            Column("RACACOR", pl.UInt8),
            Column("ESTCIV", pl.UInt8, strict=False),
            Column("ESC", pl.UInt8, strict=False),
            Column("ESC2010", pl.UInt8, strict=False),
            Column("SERIESCFAL", pl.UInt8),
            Column("OCUP", pl.Utf8),
            Column("CODMUNRES", pl.UInt32),
            Column("LOCOCOR", pl.UInt8),
            Column("CODESTAB", pl.UInt32, strict=False),
            Column("ESTABDESCR", pl.Utf8),
            Column("CODMUNOCOR", pl.UInt32),
            Column("IDADEMAE", pl.UInt16, strict=False),
            Column("ESCMAE", pl.Int8, strict=False),
            Column("ESCMAE2010", pl.Int8, strict=False),
            Column("SERIESCMAE", pl.Int8),
            Column("OCUPMAE", pl.Utf8),
            Column("QTDFILVIVO", pl.UInt8, strict=False),
            Column("QTDFILMORT", pl.UInt8, strict=False),
            Column("GRAVIDEZ", pl.UInt8),
            Column("SEMAGESTAC", pl.UInt8),
            Column("GESTACAO", pl.UInt8, strict=False),
            Column("PARTO", pl.UInt8),
            Column("OBITOPARTO", pl.UInt8),
            Column("PESO", pl.UInt32),
            Column("TPMORTEOCO", pl.UInt8),
            Column("OBITOGRAV", pl.UInt8),
            Column("OBITOPUERP", pl.UInt8),
            Column("ASSISTMED", pl.UInt8),
            Column("EXAME", pl.UInt8),
            Column("CIRURGIA", pl.UInt8),
            Column("NECROPSIA", pl.UInt8),
            Column("LINHAA", pl.Utf8),
            Column("LINHAB", pl.Utf8),
            Column("LINHAC", pl.Utf8),
            Column("LINHAD", pl.Utf8),
            Column("LINHAII", pl.Utf8),
            Column("CAUSABAS", pl.Utf8),
            Column("CB_PRE", pl.Utf8),
            Column("COMUNSVOIM", pl.Utf8),
            DateColumn("DTATESTADO", "%d%m%Y", strict=False),
            Column("CIRCOBITO", pl.UInt8, strict=False),
            Column("ACIDTRAB", pl.UInt8),
            Column("FONTE", pl.Utf8),
            Column("NUMEROLOTE", pl.Utf8),
            Column("TPPOS", pl.Utf8),
            DateColumn("DTINVESTIG", "%d%m%Y"),
            Column("CAUSABAS_O", pl.Utf8),
            DateColumn("DTCADASTRO", "%d%m%Y"),
            Column("ATESTANTE", pl.UInt8),
            Column("STCODIFICA", pl.Utf8),
            Column("CODIFICADO", pl.Utf8),
            Column("VERSAOSIST", pl.Utf8),
            Column("VERSAOSCB", pl.Utf8),
            Column("FONTEINV", pl.Utf8),
            DateColumn("DTRECEBIM", "%d%m%Y"),
            Column("ATESTADO", pl.Utf8),
            DateColumn("DTRECORIGA", "%d%m%Y"),
            Column("CAUSAMAT", pl.Utf8),
            Column("ESCMAEAGR1", pl.UInt8),
            Column("ESCFALAGR1", pl.UInt8),
            Column("STDOEPIDEM", pl.UInt8),
            Column("STDONOVA", pl.UInt8),
            Column("DIFDATA", pl.UInt16),
            Column("NUDIASOBCO", pl.UInt16),
            Column("NUDIASOBIN", pl.UInt16),
            DateColumn("DTCADINV", "%d%m%Y"),
            Column("TPOBITOCOR", pl.UInt8),
            DateColumn("DTCONINV", "%d%m%Y"),
            Column("FONTES", pl.Utf8),
            Column("TPRESGINFO", pl.UInt8),
            Column("TPNIVELINV", pl.Utf8),
            Column("NUDIASINF", pl.UInt16),
            DateColumn("DTCADINF", "%d%m%Y"),
            Column("MORTEPARTO", pl.UInt8),
            DateColumn("DTCONCASO", "%d%m%Y"),
            Column("FONTESINF", pl.Utf8),
            Column("ALTCAUSA", pl.UInt8),
            Column("CONTADOR", pl.UInt32),
        ],
    )



import polars as pl
import logging
from ..pl_utils import to_schema, Column, DateColumn
from ..datasus import import_from_ftp
from ..utils import format_year
from ..ftp import fetch_dbc_as_df

MAIN_TABLE = "PO"


def import_po(db_file="datasus.db", years=["*"]):
    """Import PO (Painel de Oncologia) data (since 2013).

    Args:
        db_file (str, optional): path to the duckdb file in which the data will be imported to. Defaults to "datasus.db".
        years (list, optional): list of years for which data will be imported (if available). Eg: `[2013, 2020]` Defaults to ["*"].

    ---

    Extra:
    - **Data description**: https://github.com/mymatsubara/datasus-db/blob/main/docs/po.pdf
    - **ftp path**: ftp.datasus.gov.br/dissemin/publicos/IBGE/POP/POPBR*.zip
    """
    logging.info(f"⏳ [{MAIN_TABLE}] Starting import...")

    import_from_ftp(
        [MAIN_TABLE],
        [
            f"/dissemin/publicos/PAINEL_ONCOLOGIA/DADOS/POBR{format_year(year, digits=4)}.dbc*"
            for year in years
        ],
        fetch_po,
        db_file=db_file,
    )


def fetch_po(ftp_path: str):
    df = fetch_dbc_as_df(ftp_path)
    return {MAIN_TABLE: map_po(df)}


def map_po(df: pl.DataFrame):
    df = df.with_columns(
        [
            pl.when(pl.col(pl.Utf8).str.len_chars() == 0)
            .then(None)
            .otherwise(pl.col(pl.Utf8))
            .name.keep(),
        ]
    )
    return to_schema(
        df,
        [
            Column("ANO_DIAGN", pl.UInt16),
            Column("ANOMES_DIA", pl.UInt32),
            Column("ANO_TRATAM", pl.UInt16),
            Column("ANOMES_TRA", pl.UInt32),
            Column("UF_RESID", pl.UInt8),
            Column("MUN_RESID", pl.UInt32),
            Column("UF_TRATAM", pl.UInt8),
            Column("MUN_TRATAM", pl.UInt32),
            Column("UF_DIAGN", pl.UInt8),
            Column("MUN_DIAG", pl.UInt32),
            Column("TRATAMENTO", pl.UInt8),
            Column("DIAGNOSTIC", pl.UInt8),
            Column("IDADE", pl.UInt8, strict=False),
            Column("SEXO", pl.Utf8),
            Column("ESTADIAM", pl.UInt8),
            Column("CNES_DIAG", pl.UInt32),
            Column("CNES_TRAT", pl.UInt32),
            Column("TEMPO_TRAT", pl.Utf8),
            Column("CNS_PAC", pl.Utf8),
            Column("DIAG_DETH", pl.Utf8),
            DateColumn("DT_DIAG", "%d/%m/%Y"),
            DateColumn("DT_TRAT", "%d/%m/%Y"),
            DateColumn("DT_NASC", "%d/%m/%Y"),
        ],
    )



"""
Module with functions to deal with DATASUS convetion files (*.cnv), which are usually file which maps ids to a readable names.
"""

import io
import re
import polars as pl
from .pl_utils import Column, to_schema


def to_dataframe(cnv_bytes: bytes, id_dtype=pl.UInt32):
    df = pl.DataFrame(
        parse_from_bytes(cnv_bytes, encoding="latin-1"),
    ).rename({"column_0": "ID", "column_1": "NOME"})

    return to_schema(df, [Column("ID", id_dtype), Column("NOME", pl.Utf8)])


def parse_from_bytes(
    bytes: bytes,
    encoding="utf-8",
    skip_rows=1,
    delimiter=r"\s{2,}",
    id_idx=3,
    label_idx=2,
):
    file = io.TextIOWrapper(io.BytesIO(bytes), encoding=encoding)
    return parse(
        file,
        skip_rows=skip_rows,
        delimiter=delimiter,
        id_idx=id_idx,
        label_idx=label_idx,
    )


def parse(
    file: io.TextIOWrapper, skip_rows=1, delimiter=r"\s{2,}", id_idx=3, label_idx=2
):
    for i, row in enumerate(file):
        if i < skip_rows:
            continue

        split = re.split(delimiter, row)
        id, label = split[id_idx].split(",")[0], split[label_idx]
        yield (id, label)



import polars as pl
import os.path as path
import logging
from ..pl_utils import to_schema, Column
from ..views.ibge_piramide_etaria import create_piramide_etaria_view
from ..datasus import import_from_ftp
from ..utils import format_year
from ..ftp import fetch_from_zip

MAIN_TABLE = "IBGE_POP"


def import_ibge_pop(db_file="datasus.db", years=["*"]):
    """Import IBGE population by age and sex per city.
    https://google.com

    Args:
        db_file (str, optional): path to the duckdb file in which the data will be imported to. Defaults to "datasus.db".
        years (list, optional): list of years for which data will be imported (if available). Eg: `[2012, 2000, 2010]. Defaults to ["*"].

    ---

    Extra:
    - **Data description**: https://github.com/mymatsubara/datasus-db/blob/main/docs/ibge_pop.pdf
    - **ftp path**: ftp.datasus.gov.br/dissemin/publicos/IBGE/POP/POPBR*.zip
    """
    logging.info(f"⏳ [{MAIN_TABLE}] Starting import...")

    import_from_ftp(
        [MAIN_TABLE],
        [
            f"/dissemin/publicos/IBGE/POP/POPBR{format_year(year)}.zip*"
            for year in years
        ],
        fetch_ibge_pop,
        db_file=db_file,
    )

    create_piramide_etaria_view()


def fetch_ibge_pop(ftp_path: str):
    csv_file = path.basename(ftp_path).split(".")[0] + ".csv"
    files = fetch_from_zip(ftp_path, [csv_file])
    df = pl.read_csv(
        files[csv_file],
        schema={
            "MUNIC_RES": pl.UInt32,
            "ANO": pl.UInt32,
            "SEXO": pl.UInt32,
            "SITUACAO": pl.UInt32,
            "FXETARIA": pl.Utf8,
            "POPULACAO": pl.UInt32,
        },
    )

    return {MAIN_TABLE: map_ibge_pop(df)}


def map_ibge_pop(df: pl.DataFrame):
    df = (
        df.with_columns(
            pl.when(pl.col("FXETARIA").is_in(["I000", "R000"]))
            .then("-100")
            .otherwise(pl.col("FXETARIA"))
            .name.keep(),
        )
        .with_columns(
            pl.col("FXETARIA").cast(pl.Int64),
        )
        .with_columns(
            pl.when(pl.col("FXETARIA") == 0)
            .then(0)
            .otherwise(pl.col("FXETARIA") // 100)
            .alias("INICIO_FXETARIA"),
            pl.when(pl.col("FXETARIA") == 0)
            .then(0)
            .otherwise(pl.col("FXETARIA") % 100)
            .alias("FIM_FXETARIA"),
        )
    )

    return to_schema(
        df,
        [
            Column("MUNIC_RES", pl.UInt32),
            Column("ANO", pl.UInt16),
            Column("SEXO", pl.UInt8),
            Column("SITUACAO", pl.UInt8),
            Column("INICIO_FXETARIA", pl.Int8),
            Column("FIM_FXETARIA", pl.Int8),
            Column("POPULACAO", pl.UInt32),
        ],
    )



"""
Module with helper functions to work with polars dataframes.
"""

import polars as pl
from dataclasses import dataclass


@dataclass
class Column:
    name: str
    dtype: pl.PolarsDataType
    strict: bool = True

    def upsert(self, df: pl.DataFrame):
        return upsert_column(df, self.name, self.dtype, strict=self.strict)


@dataclass
class DateColumn:
    name: str
    format: str
    strict: bool = True

    def upsert(self, df: pl.DataFrame):
        return upsert_date_column(df, self.name, self.format, strict=self.strict)


def upsert_column(df: pl.DataFrame, name: str, dtype: pl.PolarsDataType, strict=True):
    if name in df.columns:
        return pl.col(name).cast(dtype, strict=strict)
    else:
        return pl.lit(None, dtype).alias(name)


def upsert_date_column(df: pl.DataFrame, name: str, format: str, strict=True):
    dtype = pl.Date
    if name in df.columns:
        return pl.col(name).str.to_date(format, strict=strict)
    else:
        return pl.lit(None, dtype).alias(name)


def to_schema(df: pl.DataFrame, schema: list[Column]):
    schema_cols = {col.name for col in schema}
    cols_to_remove = [col for col in df.columns if not col in schema_cols]
    df = df.drop(cols_to_remove)

    return df.with_columns([col.upsert(df) for col in schema])


def rename_columns(df: pl.DataFrame, mapping: dict[str, str]):
    cur_cols = set(df.columns)
    mapping_possible = {
        col: mapping[col] for (col) in mapping.keys() if col in cur_cols
    }

    if len(mapping_possible) == 0:
        return df
    else:
        return df.rename(mapping_possible)


def fill_empty(fill: object, col=pl.col(pl.Utf8)):
    return pl.when(col.str.len_chars() == 0).then(fill).otherwise(col).name.keep()


def fill_text(match: str, fill: object, col=pl.col(pl.Utf8)):
    return pl.when(col == match).then(fill).otherwise(col).name.keep()


def fill_non_numeric(fill: object, col: pl.Expr):
    return pl.when(col.str.contains(r"\D")).then(fill).otherwise(col).name.keep()



"""
Module with common functions used to interact with DuckDB
"""

import duckdb
import os.path as path
import polars as pl

IMPORT_TABLE = "__import"


def create_import_table(db_con: duckdb.DuckDBPyConnection):
    db_con.sql(
        f"""
CREATE TABLE IF NOT EXISTS {IMPORT_TABLE} (
    file VARCHAR(255),
    table_name VARCHAR(255),
    created_at TIMESTAMP DEFAULT current_timestamp,
    PRIMARY KEY (file, table_name)
)"""
    )


def check_new_files(
    files: list[str], target_tables: list[str], db_con: duckdb.DuckDBPyConnection
):
    tables = ",".join((f"'{table}'" for table in target_tables))

    imported_files = db_con.query(
        f"""
SELECT file, count(*) as count 
FROM {IMPORT_TABLE}  
WHERE table_name IN ({tables})
GROUP BY file
HAVING count = {len(target_tables)}"""
    ).pl()["file"]
    imported_files = set(imported_files)

    return [file for file in files if not path.basename(file) in imported_files]


def is_file_imported(
    file: str, target_table: str, db_con: duckdb.DuckDBPyConnection
) -> bool:
    return (
        db_con.execute(
            f"SELECT COUNT(*) as count FROM {IMPORT_TABLE} WHERE table_name = ? AND file = ?",
            [target_table, path.basename(file)],
        ).pl()["count"][0]
        == 1
    )


def import_dataframe(
    table_name: str, df: pl.DataFrame, db_con: duckdb.DuckDBPyConnection
):
    # Since this is function is running on a controlled environment we don't sanitize the table name
    if has_table(table_name, db_con):
        cols = ",".join((f'"{col}"' for col in df.columns))
        db_con.sql(f"INSERT INTO {table_name} ({cols}) SELECT * FROM df")
    else:
        db_con.sql(f"CREATE TABLE {table_name} AS SELECT * FROM df")


def mark_file_as_imported(
    file: str, table_name: str, db_con: duckdb.DuckDBPyConnection
):
    db_con.execute(
        f"INSERT INTO {IMPORT_TABLE} (file, table_name) VALUES (?, ?)",
        [path.basename(file), table_name],
    )


def has_table(table_name: str, db_con: duckdb.DuckDBPyConnection) -> bool:
    return db_con.execute(
        "SELECT count(*) == 1 as has_table FROM duckdb_tables where table_name = ?",
        [table_name],
    ).pl()["has_table"][0]
